package entity

import "fmt"
import "strings"
import "github.com/alecthomas/participle/v2/lexer"

// Signature is a function or method signature that is used in functions,
// methods, and specifying interface behaviors. It defines the type of a
// function.
type Signature struct {
	// Syntax
	Pos lexer.Position
	Name      string         `parser:" '[' @Ident  "`
	Arguments []*Declaration `parser:"     @@* ']' "`
	Return    Type           `parser:" ( ':' @@ )? "`

	// Semantics
	ArgumentOrder []string
	ArgumentMap map[string] *Declaration
}
func (*Signature) ty(){}
func (this *Signature) String () string {
	out := "[" + this.Name
	for _, argument := range this.Arguments {
		out += fmt.Sprint(" ", argument)
	}
	out += "]"
	if this.Return != nil {
		out += fmt.Sprint(":", this.Return)
	}
	return out
}
func (this *Signature) Equals (ty Type) bool {
	real, ok := ty.(*Signature)
	if !ok ||
		len(real.Arguments) != len(this.Arguments) ||
		!real.Return.Equals(this.Return) ||
		real.Name != this.Name { return false }

	for index, argument := range this.Arguments {
		if !argument.Type.Equals(real.Arguments[index].Type) {
			return false
		}
	}
	return true
}

// Member is a syntactical construct that is used to list members in struct
// literals.
type Member struct {
	Pos lexer.Position
	Name  string     `parser:" @Ident "`
	Value Expression `parser:" ':' @@ "`
}
func (this *Member) String () string {
	return fmt.Sprint(this.Name, ":", this.Value)
}

// Operator determines which operation to apply to a value in an operation
// expression.
type Operator int; const (
	// Math
	OperatorAdd Operator = iota
	OperatorIncrement
	OperatorSubtract
	OperatorDecrement
	OperatorMultiply
	OperatorDivide
	OperatorModulo

	// Logic
	OperatorLogicalNot
	OperatorLogicalOr
	OperatorLogicalAnd
	OperatorLogicalXor

	// Bit manipulation
	OperatorNot
	OperatorOr
	OperatorAnd
	OperatorXor
	OperatorLeftShift
	OperatorRightShift

	// Comparison
	OperatorLess
	OperatorGreater
	OperatorLessEqual
	OperatorGreaterEqual
	OperatorEqual
)

var operatorToString = []string {
	// Math
	"+",
	"++",
	"-",
	"--",
	"*",
	"/",
	"%",

	// Logic
	"!",
	"|",
	"&",
	"^",
	// Bit manipulation
	"!!",
	"||",
	"&&",
	"^^",
	"<<",
	">>",

	// Comparison
	"<",
	">",
	"<=",
	">=",
	"=",
}

var stringToOperator map[string] Operator

func init () {
	stringToOperator = make(map[string] Operator)
	for index, str := range operatorToString {
		stringToOperator[str] = Operator(index)
	}
}

func (this *Operator) Capture (str []string) error {
	*this = stringToOperator[strings.Join(str, "")]
	return nil
}

func (this Operator) String () string {
	return operatorToString[this]
}
